@@ -1,6 +1,6 @@
 {-
-Bustle.Diagram: turn a list of messages into a set of abstract shapes
-Copyright (C) 2008 Collabora Ltd.
+Bustle.Diagram: shapes for sequence diagrams
+Copyright (C) 2008–2009 Collabora Ltd.
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
@@ -18,23 +18,33 @@ Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 -}
 module Bustle.Diagram
   ( Shape(..)
+  , Diagram
   , Arrowhead(..)
   , Side(..)
   , Colour(..)
   , Rect
-  , bounds
-  , draw
-  , clearCanvas
-  , drawBoundingBox
-  , intersects
+  , dimensions
+  , topJustify
+  , drawDiagram
+  , drawRegion
+  , headers
+  , headerHeight
+  , columnWidth
   )
 where
 
 import Data.Maybe (maybe)
+import Data.List (unzip4)
 import Control.Arrow ((&&&))
 import Control.Applicative ((<$>), (<*>))
+import Control.Monad (forM_)
+
+import Control.Monad.Reader
 
 import Graphics.Rendering.Cairo
+import Graphics.UI.Gtk.Cairo (cairoCreateContext, showLayout)
+import Graphics.UI.Gtk.Pango.Layout
+import Graphics.UI.Gtk.Pango.Font
 
 type Point = (Double, Double)
 type Rect = (Double, Double, Double, Double)
@@ -61,9 +71,12 @@ offset R = (+)
 data Colour = Colour Double Double Double
   deriving (Eq, Show, Read, Ord)
 
-data Shape = Header { str :: String, shapex, shapey :: Double
+data Shape = Header { strs :: [String]
+                    , shapex, shapey :: Double
                     }
-           | MemberLabel String String Double
+           | MemberLabel { labelPath, labelMember :: String
+                         , shapey :: Double
+                         }
            | Timestamp { str :: String, shapey :: Double }
            | ClientLine { shapex, shapey1, shapey2 :: Double }
            | Rule { shapex, shapey :: Double }
@@ -78,6 +91,8 @@ data Shape = Header { str :: String, shapex, shapey :: Double
                  }
   deriving (Show, Read, Eq)
 
+type Diagram = [Shape]
+
 arcControlPoints :: Shape -> (Point, Point)
 arcControlPoints (Arc { topx=x1, topy=y1, bottomx=x2, bottomy=y2, arcside=s }) =
     let (+-) = offset s
@@ -86,6 +101,15 @@ arcControlPoints (Arc { topx=x1, topy=y1, bottomx=x2, bottomy=y2, arcside=s }) =
     in (cp1, cp2)
 arcControlPoints _ = error "i see you've played arcy-shapey before"
 
+mapY :: (Double -> Double) -> (Shape -> Shape)
+mapY f s = case s of
+    Arc {}        -> s { topy = f (topy s)
+                       , bottomy = f (bottomy s)
+                       }
+    ClientLine {} -> s { shapey1 = f (shapey1 s)
+                       , shapey2 = f (shapey2 s)
+                       }
+    _             -> s { shapey = f (shapey s) }
 
 --
 -- Constants
@@ -102,7 +126,7 @@ memberx = 230
 memberWidth = 340
 
 columnWidth :: Double
-columnWidth = 70
+columnWidth = 90
 
 
 --
@@ -120,6 +144,9 @@ fromCentre x y width =
      x + width / 2, y + height / 2)
   where height = eventHeight
 
+headerHeight :: [String] -> Double
+headerHeight = fromIntegral . (10 *) . length
+
 bounds :: Shape -> Rect
 bounds s = case s of
   ClientLine {} -> (shapex s, shapey1 s, shapex s, shapey2 s)
@@ -135,20 +162,65 @@ bounds s = case s of
     in (x1, y1, x2, y2)
   Arc { topx=x1, bottomx=x2, topy=y1, bottomy=y2 } ->
     let ((cx, _), (dx, _)) = arcControlPoints s
-    in (min x1 cx, y1, max x2 dx, y2)
+       -- FIXME: magic 5 makes the bounding box include the text
+    in (min x1 cx, y1, max x2 dx, y2 + 5)
   Timestamp { shapey=y } -> fromCentre timestampx y timestampWidth
   MemberLabel _ _ y -> fromCentre memberx y memberWidth
-  Header { shapex = x, shapey = y} -> fromCentre x y columnWidth
-
+  Header { strs = ss, shapex = x, shapey = y} ->
+    let width = columnWidth
+        height = headerHeight ss
+    in (x - width / 2, y,
+        x + width / 2, y + height)
 
 intersects :: Rect -> Rect -> Bool
 intersects (x,y,w,z) (x', y', w', z') =
   not $ or [x > w', w < x', y > z', z < y']
 
+-- Constructs a series of headers of various-sized lists of names,
+-- bottom-justified.
+headers :: [(Double, [String])]  -- list of (x-coordinate, names)
+        -> Double                -- y-coordinate of top of headers
+        -> (Double, [Shape])     -- the headers' combined height, and shapes
+headers []  _ = (0, [])
+headers xss y = (height, shapes)
+  where heights = map (headerHeight . snd) xss
+        height  = maximum heights
+        adjs    = map (height -) heights
+        shapes  = zipWith (\(x, ss) adj -> Header ss x (y + adj)) xss adjs
+
 --
 -- Drawing
 --
 
+dimensions :: Diagram -> (Double, Double)
+dimensions shapes = (x2, y2)
+    where (_, _, x2, y2) = corners shapes
+
+corners :: Diagram -> (Double, Double, Double, Double)
+corners [] = (0, 0, 0, 0)
+corners shapes = (minimum x1s, minimum y1s,
+                  maximum x2s, maximum y2s)
+    where (x1s, y1s, x2s, y2s) = unzip4 $ map bounds shapes
+
+topJustify :: Diagram -> Diagram
+topJustify shapes = f shapes
+    where (_, y1, _, _) = corners shapes
+          f             = if y1 < 0 then map (mapY (subtract y1)) else id
+
+drawDiagram :: Bool -> Diagram -> Render ()
+drawDiagram drawBounds shapes = do
+    clearCanvas
+    forM_ shapes $ \x -> do
+        when drawBounds (drawBoundingBox x)
+        draw x
+
+drawRegion :: Rect -> Bool -> Diagram -> Render ()
+drawRegion r db = drawDiagram db . visible r . map (bounds &&& id)
+
+visible :: Rect -> [(Rect, Shape)] -> [Shape]
+visible r = map snd . filter (intersects r . fst)
+
+
 saved :: Render () -> Render ()
 saved act = save >> act >> restore
 
@@ -175,7 +247,7 @@ draw s = draw' s
                               shapex2 <*> shapey
           Arrow {} -> drawArrow <$> shapecolour <*> arrowhead <*> shapex1 <*>
                         shapex2 <*> shapey
-          Header {} -> drawHeader <$> str <*> shapex <*> shapey
+          Header {} -> drawHeader <$> strs <*> shapex <*> shapey
           MemberLabel s1 s2 y -> const (drawMember s1 s2 y)
           Timestamp {} -> drawTimestamp <$> str <*> shapey
           ClientLine {} -> drawClientLine <$> shapex <*> shapey1 <*> shapey2
@@ -224,9 +296,7 @@ drawArc :: Double -> Double -> Double -> Double
         -> Double -> Double -> Double -> Double
         -> String
         -> Render ()
-drawArc cx cy dx dy x1 y1 x2 y2 cap = do
-    save
-
+drawArc cx cy dx dy x1 y1 x2 y2 cap = saved $ do
     setSourceRGB 0.4 0.7 0.4
     setDash [3, 3] 0
 
@@ -235,41 +305,56 @@ drawArc cx cy dx dy x1 y1 x2 y2 cap = do
     stroke
 
     setSourceRGB 0 0 0
-    extents <- textExtents cap
-    let textWidth = textExtentsWidth extents
-        tx = min x2 dx + abs (x2 - dx) / 2
-    moveTo (if x1 > cx then tx - textWidth else tx) y2
-    showText cap
+    l <- mkLayout cap EllipsizeNone AlignLeft
+    (PangoRectangle _ _ textWidth _, _) <- liftIO $ layoutGetExtents l
+    let tx = min x2 dx + abs (x2 - dx) / 2
+    moveTo (if x1 > cx then tx - textWidth else tx) (y2 - 5)
+    showLayout l
 
-{-
-    -- bounding box of arc; useful as a guide to the coordinates we have to
-    -- play with.
-    setSourceRGB 0 0 1
-    rectangle x1 y1 (cx - x1) (y2 - y1)
-    stroke
--}
+font :: IO FontDescription
+font = do
+    fd <- fontDescriptionNew
+    fontDescriptionSetSize fd 7
+    fontDescriptionSetFamily fd "Sans"
+    return fd
 
-    restore
+mkLayout :: (MonadIO m)
+         => String -> EllipsizeMode -> LayoutAlignment
+         -> m PangoLayout
+mkLayout s e a = liftIO $ do
+    ctx <- cairoCreateContext Nothing
+    layout <- layoutEmpty ctx
+    layoutSetMarkup layout s
+    layoutSetFontDescription layout . Just =<< font
+    layoutSetEllipsize layout e
+    layoutSetAlignment layout a
+    return layout
 
-drawHeader :: String -> Double -> Double -> Render ()
-drawHeader name x y = do
-    extents <- textExtents name
-    let diff = textExtentsWidth extents / 2
-    moveTo (x - diff) (y + 10)
-    showText name
+withWidth :: MonadIO m => m PangoLayout -> Double -> m PangoLayout
+withWidth m w = do
+    l <- m
+    liftIO $ layoutSetWidth l (Just w)
+    return l
+
+drawHeader :: [String] -> Double -> Double -> Render ()
+drawHeader names x y = forM_ (zip [0..] names) $ \(i, name) -> do
+    l <- mkLayout name EllipsizeEnd AlignCenter `withWidth` columnWidth
+    moveTo (x - (columnWidth / 2)) (y + i * h)
+    showLayout l
+  where h = 10
 
 drawMember :: String -> String -> Double -> Render ()
-drawMember s1 s2 y = do
-    -- TODO: Make this fit in the bounds we estimated
-    moveTo timestampWidth y
-    showText s1
-    moveTo timestampWidth (y + 10)
-    showText s2
+drawMember s1 s2 y = dm s1 (y - 10) >> dm s2 y
+  where
+    dm s y' = do
+      l <- mkLayout s EllipsizeStart AlignLeft `withWidth` memberWidth
+      moveTo timestampWidth y'
+      showLayout l
 
 drawTimestamp :: String -> Double -> Render ()
 drawTimestamp ts y = do
-    moveTo 0 y
-    showText ts
+    moveTo 0 (y - 10)
+    showLayout =<< mkLayout ts EllipsizeNone AlignLeft `withWidth` timestampWidth
 
 drawClientLine :: Double -> Double -> Double -> Render ()
 drawClientLine x y1 y2 = saved $ do
